# Copyright 2019, Oath Inc.
# Licensed under the terms of the BSD 3 Clause license. See LICENSE file in project root for terms.
"""
pypirun module
"""
from typing import List
import pkg_resources


__all__: List[str] = []
__copyright__: str = "Copyright 2019, Oath Inc."
__version__: str = pkg_resources.get_distribution("pypirun").version



# Copyright 2019, Oath Inc.
# Licensed under the terms of the BSD 3 Clause license. See LICENSE file in project root for terms.
"""
pypirun utility functions
"""
import os
import shutil


def env_bool(variable_name, default=True):
    """
    Interperate an environment variable as a boolean value.

    Parameters
    ==========
    variable_name: str
        Environment key to get the value for

    default: bool
        Value to return if the environment variable does not exists

    Returns
    =======
    bool: Value from the environment variable
    """
    true_strings = ['true', 'on', '1']

    value = os.environ.get(variable_name, str(default)).lower()

    if value in true_strings:
        return True
    return False


def which(filename, allow_symlink=False):
    """
    Search the path for an executable, optionally ignoring symlinks

    Parameters
    ==========
    filename: str
        Filename to search for

    allow_symlink: bool
        If True, return filenames that are symlinks
    """
    if allow_symlink:  # pragma: no cover
        return shutil.which(filename)

    search_dirs = os.environ.get('PATH', '').split(':')
    for directory in search_dirs:  # pragma: no cover
        full_filename = os.path.join(directory, filename)
        if not os.path.exists(full_filename):
            continue
        if os.path.islink(full_filename):
            link = os.readlink(full_filename)
            if '/' in link:
                return os.path.abspath((os.path.join(directory, link)))
            continue
        return full_filename



# Copyright 2019, Oath Inc.
# Licensed under the terms of the BSD 3 Clause license. See LICENSE file in project root for terms.
"""
pypirun command line utility
"""
import os
import shutil
import subprocess  # nosec
import sys
import tempfile
from .arguments import parse_arguments, ParseError
from .utility import which


def interpreter_version(interpreter):
    """
    Get the version from the python interpreter
    """
    version = subprocess.check_output([interpreter, '--version'], stderr=subprocess.STDOUT).decode(errors='ignore').strip().split()[-1]  # nosec
    return version


def interpreter_parent(interpreter):
    """
    Get the parent python interpreter for interpreter (I.E the interpreter that created the virtualenv if it is an
    interpreter in a virtualenv.

    Returns
    =======
    str:
        parent interpreter
    """
    try:
        real_prefix = subprocess.check_output([interpreter, '-c', 'import sys;print(sys.real_prefix)'], stderr=subprocess.DEVNULL).decode(errors='ignore').strip()  # nosec
    except subprocess.CalledProcessError:  # pragma: no cover
        return interpreter
    try:  # pragma: no cover
        major_minor = subprocess.check_output([interpreter, '-c', 'import sys;print(f"{sys.version_info.major}.{sys.version_info.minor}")'], stderr=subprocess.DEVNULL).decode(errors='ignore').strip()  # nosec
        basename = f'python{major_minor}'
    except subprocess.CalledProcessError:  # pragma: no cover
        basename = 'python3'

    bin_dir = os.path.join(real_prefix, 'bin')
    interpreter = os.path.join(bin_dir, basename)
    return interpreter


def install_and_run(package, command, interpreter, debug=False, no_cache_dir=False, upgrade_setuptools=False, upgrade_pip=False):
    """
    Install a package and run a command in a temporary Python virtualenv
    
    Parameters
    ==========
    package: str
        A string containing a comma seperated list of packages to install.
    
    command: str
        The command to run in the shell once the package is installed
        
    interpreter: str
        The python interpreter executable to use to create the virtualenv
    
    debug: bool, optional
        Print more useful debug output.  Default: False
    
    no_cache_dir: bool, optional
        If True, pass the no-cache-dir flag to the pip command to disable pip package caching.  Default: False
    
    upgrade_setuptools: bool, optional
        Upgrade setuptools after creating the virtualenv but before installing packages.  Default: False

    upgrade_pip: bool, optional
        Upgrade pip after creating the virtualenv but before installing packages.  Default: False
    """
    packages = package.split(',')

    with tempfile.TemporaryDirectory() as tempdir:
        venv_dir = os.path.join(tempdir, '.venv')
        venv_bin = os.path.join(venv_dir, 'bin')
        venv_python = os.path.join(venv_bin, 'python')
        venv_pip = os.path.join(venv_bin, 'pip')

        pip_args = []
        if no_cache_dir:  # pragma: no cover
            pip_args = ['--no-cache-dir']

        # Create venv
        try:
            output = subprocess.check_output([interpreter, '-m', 'venv', venv_dir])  # nosec
        except subprocess.CalledProcessError:  # pragma: no cover
            print('Failed to create temporary virtualenv using the', interpreter, 'python interpreter')
            return 1
        if debug:
            if output.decode().strip():  # pragma: no cover
                print(output.decode().strip())

        if not os.path.exists(venv_pip):  # pragma: no cover
            output = subprocess.check_output([venv_python, '-m', 'pip', 'install', '--force-reinstall'] + pip_args + ['pip'], stderr=subprocess.STDOUT)  # nosec
            if debug:  # pragma: no cover
                print(output.decode())

        if upgrade_pip:  # pragma: no cover
            output = subprocess.check_output([venv_pip, 'install', '--upgrade'] + pip_args + ['pip'], stderr=subprocess.STDOUT)  # nosec
            if debug:  # pragma: no cover
                print(output.decode())

        if upgrade_setuptools:  # pragma: no cover
            output = subprocess.check_output([venv_pip, 'install', '--upgrade'] + pip_args + ['setuptools'], stderr=subprocess.STDOUT)  # nosec
            if debug:  # pragma: no cover
                print(output.decode())

        venv_bin_before = set(os.listdir(venv_bin))

        # Install the package
        try:
            output = subprocess.check_output([venv_pip, 'install'] + pip_args + packages, stderr=subprocess.STDOUT)  # nosec
        except subprocess.CalledProcessError as error:   # pragma: no cover
            print(error.output.decode(), file=sys.stdout)
            return error.returncode

        venv_bin_after = set(os.listdir(venv_bin))

        if debug:  # pragma: no cover
            print(output.decode())
            print(f'Installed files: {list(venv_bin_after-venv_bin_before)}')

        # Run the command
        try:
            subprocess.check_call(f'{venv_dir}/bin/{command}', shell=True)  # nosec
        except subprocess.CalledProcessError as error:  # pragma: no cover
            return error.returncode
        return exit_ok()  # pragma: no cover


def exit_ok():
    """A python implementation of /bin/true that can be used for testing"""
    return 0


def main():
    """
    Command line interface entrypoint
    
    Returns
    =======
    int:
        return code from running the command
    """
    try:
        args = parse_arguments()
    except ParseError:
        return 1
    interpreter = args.interpreter
    command_file = args.command[0]
    command = ' '.join(args.command)
    if not interpreter:    # pragma: no cover
        interpreter = interpreter_parent(sys.executable)
    if not interpreter:    # pragma: no cover
        interpreter = which('python3')
    if not interpreter:    # pragma: no cover
        print('Unable to find python3 interpreter')
        return 1

    if args.always_install or not shutil.which(command_file):
        return install_and_run(package=args.package, command=command, interpreter=interpreter, debug=args.debug, no_cache_dir=args.no_cache_dir, upgrade_setuptools=args.upgrade_setuptools, upgrade_pip=args.upgrade_pip)

    try:
        subprocess.check_call(command, shell=True)  # nosec
    except subprocess.CalledProcessError as error:
        return error.returncode
    return 0  # pragma: no cover



# Copyright 2019, Oath Inc.
# Licensed under the terms of the BSD 3 Clause license. See LICENSE file in project root for terms.
"""
pypirun module command line parsing functions
"""
import argparse
import os
import sys
from .utility import env_bool


class ParseError(Exception):
    """
    CLI Argument parsing error
    """
    pass


def parse_arguments() -> argparse.Namespace:
    """
    Parse the command line arguments

    Returns
    =======
    argparse.Namespace:
        Arguments parsed from the command line

    Raises
    ======
    ParseError - Unable to parse the command line arguments
    """

    default_interpreter = os.environ.get('BASE_PYTHON', None)
    debug = env_bool('PYPIRUN_DEBUG', False)

    parser = argparse.ArgumentParser()
    parser.add_argument('--interpreter', default=default_interpreter, help='Python interpreter to use')
    parser.add_argument('--debug', default=debug, action='store_true', help='Enable debug output')
    parser.add_argument('--always_install', default=False, action='store_true', help='Install the command even if it exists in the path')
    parser.add_argument('--no-cache-dir', default=False, action='store_true', help="Disable the pip cache when installing")
    parser.add_argument('--upgrade_pip', default=False, action='store_true', help='Upgrade the pip before installing packages')
    parser.add_argument('--upgrade_setuptools', default=False, action='store_true', help="Upgrade setuptools before installing packages")
    parser.add_argument('package', type=str, help='Package the command is in')
    parser.add_argument('command', nargs='*', help='Command to run')

    if len(sys.argv) < 2:
        parser.print_usage()
        raise ParseError('Unable to parse arguments')

    argv = []
    command = []
    in_command = False
    for argument in sys.argv[1:]:
        if not in_command and argument.startswith('--'):
            argv.append(argument)
            continue
        in_command = True
        command.append(argument)

    argv += ['package', 'command']

    if len(command) < 2:
        parser.print_usage()
        raise ParseError('Insufficient arguments provided')

    args = parser.parse_args(args=argv)
    args.package = command[0]
    args.command = command[1:]

    return args



#!/usr/bin/env python
# Copyright 2019, Oath Inc.
# Licensed under the terms of the BSD 3 Clause license. See LICENSE file in project root for terms.
import setuptools
import sys


def setuptools_version_supported():
    major, minor, patch = setuptools.__version__.split('.')
    if int(major) > 31:
        return True
    return False
        

if __name__ == '__main__':
    if not setuptools_version_supported():
        print('Setuptools version 31.0.0 or higher is needed to install this package')
        sys.exit(1)

    # We're being run from the command line so call setup with our arguments
    setuptools.setup()
